import numpy as np
import itertools
import threading
from typing import Callable, Union, Optional, List, Iterable

import _feyn
import feyn

from math import log


def fit_models(
    models: List[feyn.Model],
    data,
    loss_function: Union[str, Callable] = _feyn.DEFAULT_LOSS,
    criterion: Optional[str] = None,
    n_samples: Optional[int] = None,
    sample_weights: Optional[Iterable[float]] = None,
    threads: int = 4,
    immutable: bool = False,
) -> List[feyn.Model]:
    """Fit a list of models on some data and return a list of fitted models. The return list will be sorted in ascending order by either the loss function or one of the criteria.

    The n_samples parameter controls how many samples are used to train each model. The default behavior is to fit each model once with each sample in the dataset, unless the set is smaller than 10000, in which case the dataset will be upsampled to 10000 samples before fitting.

    The samples are shuffled randomly before fitting to avoid issues with the Stochastic Gradient Descent algorithm.


    Arguments:
        models {List[feyn.Model]} -- A list of feyn models to be fitted.
        data {[type]} -- Data used in fitting each model.

    Keyword Arguments:
        loss_function {Union[str, Callable]} -- The loss function to optimize models for. Can take any loss function in `feyn.losses`. (default: {"squared_error"} (MSE))
        criterion {Optional[str]} -- Sort by information criterion rather than loss. Either "aic", "bic" or None. (default: {None})
        n_samples {Optional[int]} -- The number of samples to fit each model with. (default: {None})
        sample_weights {Optional[Iterable[float]]} -- An optional numpy array of weights for each sample. If present, the array must have the same size as the data set, i.e. one weight for each sample. (default: {None})
        threads {int} -- Number of concurrent threads to use for fitting. (default: {4})
        immutable {bool} -- If True, create a copy of each model and fit those, leaving the originals untouched. This increases runtime. (default: {False})

    Raises:
        ValueError: 0 samples for fitting, sizes of data and sample weights don't match, invalid loss function.

    Returns:
        List[feyn.Model] -- A list of fitted feyn models.
    """

    if immutable:
        models = [feyn.Model._from_dict(m._to_dict(include_state=True)) for m in models]

    # Magic support for pandas DataFrame
    if type(data).__name__ == "DataFrame":
        data = {col: data[col].values for col in data.columns}

    length = len(next(iter(data.values())))

    if n_samples is None:
        n_samples = max(10000, length)
    if n_samples <= 0:
        raise ValueError("More than 0 samples are required for fitting models.")

    # Create a sequence of indices from the permutated data of length n_samples
    permutation = np.random.permutation(n_samples) % length
    data = {key: values[permutation] for key, values in data.items()}

    if sample_weights is not None:
        # Normalise the sample_weights
        sample_weights = sample_weights / max(sample_weights)
        s_size = len(sample_weights)
        if not s_size == length:
            raise ValueError(
                f"The sizes of data ({length}) and the sample weights ({s_size}) do not match."
            )

        # Also permutate the sample_weights
        sample_weights = sample_weights[permutation]

    loss_function = feyn.losses._get_loss_function(loss_function)
    if not hasattr(loss_function, "c_derivative"):
        raise ValueError(
            "Loss function cannot be used for fitting, since it doesn't have a corresponding c derivative"
        )

    _counter = itertools.count()
    _terminate = False
    _exception: BaseException = None

    def fitting_thread():
        nonlocal _terminate, _counter, _exception
        try:
            while not _terminate:
                ix = next(_counter)
                if ix >= len(models):
                    return
                m = models[ix]
                m._fit(data, loss_function, sample_weights)

                if not np.isfinite(m.loss_value):
                    # Model not defined in its entire domain. Ignore it
                    continue

        except BaseException as e:
            _exception = e

    if threads > 1:

        threadlist = [threading.Thread(target=fitting_thread) for _ in range(threads)]
        try:
            [t.start() for t in threadlist]
            while any(t.is_alive() for t in threadlist):
                [t.join(1 / threads) for t in threadlist]
                if _exception:
                    raise _exception
        finally:
            _terminate = True
            [t.join() for t in threadlist]
        if _exception:
            raise _exception

    else:
        # Dont use threading at all if threads = 1, mainly useful for debugging and profiling
        fitting_thread()

    # Remove any models that dont work for the input domain
    models = list(filter(lambda m: np.isfinite(m.loss_value), models))

    for m in models:
        m.bic = feyn.criteria.bic(m.loss_value, m._paramcount, length)
        m.aic = feyn.criteria.aic(m.loss_value, m._paramcount, length)

    if criterion == "bic":
        models.sort(key=lambda m: m.bic, reverse=False)
    elif criterion == "aic":
        models.sort(key=lambda m: m.aic, reverse=False)
    elif criterion is None:
        models.sort(key=lambda m: m.loss_value, reverse=False)
    else:
        raise Exception(
            "Unknown information criterion %s. Must be 'aic', 'bic' or None)"
            % criterion
        )

    for m in models:
        m.age += 1

    return models
